unit hiCOMMultiTh;

interface

uses Kol,Share,Windows,Debug;

{$ifdef F_P}
//   FPC,  _COMSTAT
type
  _COMSTAT = record
    Reserved: DWORD;
    cbInQue: DWORD;
    cbOutQue: DWORD;
  end;
  TComStat = _COMSTAT;
  COMSTAT = _COMSTAT;
  PComStat = ^TComStat;
{$endif}

type
  THICOMMultiTh = class(TDebug)
   private
    _ByteSize: integer;
    _StopBits: integer;
    _Parity: integer;
    _ReadBuffer: integer;
    _InOutQueue: integer;    
    BufferRd, ReadStr, BufferWr: string;    
    Sended: DWORD;    
    hFile: THandle;
    thrd, thwr: PThread;
    OvrRd, OvrWr: TOverlapped;

    procedure CloseCom;
    procedure SetByteSize(Value: integer);
    procedure SetStopBits(Value: integer);
    procedure SetParity(Value: integer);
    
    function ExecuteRd(Sender: PThread): Integer;
    function ExecuteWr(Sender: PThread): Integer;
    
    procedure SyncExecRd;        
    procedure SyncExecWr;

   public
    _prop_Port: integer;
    _prop_BaudRate: integer;

    _event_onOpen: THI_Event;
    _event_onClose: THI_Event;    
    _event_onRead: THI_Event;
    _event_onWrite: THI_Event;
    _event_onEnumPorts: THI_Event;
    _event_onEnumPortsESV: THI_Event;
    _event_onOpenError: THI_Event;    

    _data_Port: THI_Event;
    _data_BaudRate: THI_Event;

    destructor Destroy; override;
    procedure _work_doOpen(var _Data:TData; Index:word);
    procedure _work_doClose(var _Data:TData; Index:word);
    procedure _work_doWrite(var _Data:TData; Index:word);
    procedure _work_doStartRead(var _Data:TData; Index:word);
    procedure _work_doStopRead(var _Data:TData; Index:word);
    procedure _work_doEnumPorts(var _Data:TData; Index:word);
    procedure _work_doDTR(var _Data:TData; Index:word);
    procedure _work_doRTS(var _Data:TData; Index:word);
    procedure _work_doByteSize(var _Data:TData; Index:word);
    procedure _work_doStopBits(var _Data:TData; Index:word);
    procedure _work_doParity(var _Data:TData; Index:word);
    procedure _work_doReadBuffer(var _Data:TData; Index:word);
    procedure _work_doInOutQueue(var _Data:TData; Index:word);                                
    property _prop_ByteSize: integer write SetByteSize;
    property _prop_StopBits: integer write SetStopBits;
    property _prop_Parity: integer write SetParity;
    property _prop_ReadBuffer: integer write _ReadBuffer;
    property _prop_InOutQueue: integer write _InOutQueue;       

  end;

implementation

destructor THICOMMultiTh.Destroy;
begin
  CloseCom; 
  inherited; 
end;

procedure THICOMMultiTh.CloseCom;
begin
  if Assigned(thrd) then free_and_nil(thrd);
  if Assigned(thrd) then free_and_nil(thwr);

  if hFile = INVALID_HANDLE_VALUE then exit;
  PurgeComm(hFile, PURGE_TXCLEAR or PURGE_RXCLEAR);
  CloseHandle(hFile);
  hFile := INVALID_HANDLE_VALUE;
  _hi_onEvent(_event_onClose);
end;

procedure THICOMMultiTh._work_doEnumPorts;
var
  fn: string;
  i: integer;
  sFile: THandle;
begin
  for i := 1 to 256 do
  begin
    fn := '\\.\Com' + Int2Str(i);
    sFile := CreateFile(PChar(fn), GENERIC_READ or GENERIC_WRITE, 0, nil,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if sFile <> INVALID_HANDLE_VALUE then
    begin
      if GetLastError = ERROR_SHARING_VIOLATION then
        _hi_onEvent(_event_onEnumPortsESV, i)
      else
        _hi_onEvent(_event_onEnumPorts, i);
      CloseHandle(sFile);  
    end;  
  end;   
end;

procedure THICOMMultiTh._work_doOpen;
var
  cto: _COMMTIMEOUTS;
  dcb: TDCB;  
  fn: string;
  _BaudRate: integer;
  PortParam: string;   
begin
  CloseCom;
  fn := '\\.\Com' + Int2Str(ReadInteger(_Data,_data_Port,_prop_Port + 1));
  _BaudRate := ReadInteger(_Data, _data_BaudRate, _prop_BaudRate);
  hFile := CreateFile(PChar(fn), GENERIC_READ or GENERIC_WRITE, 0, nil,
                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL or FILE_FLAG_OVERLAPPED, 0);
  if hFile <> INVALID_HANDLE_VALUE then
    if GetCommState(hFile, dcb) then    //    
    begin
      dcb.BaudRate        := _BaudRate; //   
      dcb.ByteSize        := _ByteSize; //  -   
      dcb.Parity          := _Parity;   //   
      dcb.StopBits        := _StopBits; //   

      PortParam := 'xon=off odsr=off octs=off dtr=off rts=off idsr=off';
      FillChar(cto, sizeof(cto), #0);   //   TimeOut-,       

      if (BuildCommDCB(PChar(PortParam), dcb) and SetCommState(hFile, dcb) and SetCommTimeouts(hFile, cto)) then
      begin
        thrd := {$ifdef F_P}NewThreadforFPC{$else}NewThread{$endif};
        thrd.ThreadPriority := THREAD_PRIORITY_HIGHEST;
        thrd.OnExecute := ExecuteRd;
        
        thwr := {$ifdef F_P}NewThreadforFPC{$else}NewThread{$endif};
        thwr.ThreadPriority := THREAD_PRIORITY_HIGHEST;
        thwr.OnExecute := ExecuteWr;

        SetLength(BufferRd, _ReadBuffer);

        SetupComm(hFile, _InOutQueue, _InOutQueue);
        PurgeComm(hFile, PURGE_TXCLEAR or PURGE_RXCLEAR);
        _hi_CreateEvent(_Data, @_event_onOpen);
        exit;
      end;  
    end;
  CloseCom;
  _hi_CreateEvent(_Data, @_event_onOpenError);
end;

procedure THICOMMultiTh._work_doClose;
begin
  CloseCom;
end;

procedure THICOMMultiTh._work_doWrite;
begin
  BufferWr := ToString(_Data);
  if (hFile = INVALID_HANDLE_VALUE) or (BufferWr ='') then exit;
  thwr.Resume;
end;

procedure THICOMMultiTh._work_doStartRead;
begin
  if hFile = INVALID_HANDLE_VALUE then exit;
  PurgeComm(hFile, PURGE_RXCLEAR);  
  thrd.Resume;
end;

procedure THICOMMultiTh._work_doStopRead;
begin
  if hFile = INVALID_HANDLE_VALUE then exit;
  thrd.Suspend;
end;

function THICOMMultiTh.ExecuteRd(Sender: PThread): Integer;
var
  Signaled, BytesTrans, Mask, Err: DWORD;
  FStat: TComStat;  
begin
  FillChar(OvrRd, SizeOf(TOverlapped), #0);
  OvrRd.hEvent := CreateEvent(nil, True, True, #0);
  SetCommMask(hFile, EV_RXCHAR);

  while not Sender.Terminated do
  begin
    WaitCommEvent(hFile, Mask, @OvrRd);
    Signaled := WaitForSingleObject(OvrRd.hEvent, INFINITE);
    if (Signaled = WAIT_OBJECT_0) then 
      if GetOverlappedResult(hFile, OvrRd, BytesTrans, True) then
        if ((Mask and EV_RXCHAR) <> 0) then
          if ClearCommError(hFile, Err, @FStat) then
            if (FStat.cbInQue <> 0) then
            begin
              ReadFile(hFile, BufferRd[1], FStat.cbInQue, BytesTrans, @OvrRd);
              ReadStr := BufferRd;
              SetLength(ReadStr, BytesTrans);
              Sender.Synchronize(SyncExecRd);
            end;
  end;
        
  PurgeComm(hFile, PURGE_RXCLEAR);
  CloseHandle(OvrRd.hEvent);
  Result := 0;
end;

function THICOMMultiTh.ExecuteWr(Sender: PThread): Integer;
var
  Signaled: DWORD;
begin
  FillChar(OvrWr, SizeOf(TOverlapped), #0);
  OvrWr.hEvent := CreateEvent(nil, True, True, #0);
  while not Sender.Terminated do
  begin
    WriteFile(hFile, BufferWr[1], Length(BufferWr), Sended, @OvrWr);
    Signaled := WaitForSingleObject(OvrWr.hEvent, INFINITE);
    if (Signaled = WAIT_OBJECT_0) then
      if GetOverlappedResult(hFile, OvrWr, Sended, True) then
      begin
        Sender.Synchronize(SyncExecWr);
        Sender.Suspend;
      end;
  end;

  PurgeComm(hFile, PURGE_TXCLEAR);
  CloseHandle(OvrWr.hEvent);
  Result := 0;
end;

procedure THICOMMultiTh.SyncExecRd;
begin
  _hi_onEvent(_event_onRead, ReadStr);
end;

procedure THICOMMultiTh.SyncExecWr;
begin
  _hi_onEvent(_event_onWrite, integer(Sended));
end;

procedure THICOMMultiTh._work_doDTR;
begin
  if hFile = INVALID_HANDLE_VALUE then exit;
  if ReadBool(_Data) then
    EscapeCommFunction(hFile, SETDTR)
  else
    EscapeCommFunction(hFile, CLRDTR);
end;

procedure THICOMMultiTh._work_doRTS;
begin
  if hFile = INVALID_HANDLE_VALUE then exit;
  if ReadBool(_Data) then
    EscapeCommFunction(hFile, SETRTS)
  else
    EscapeCommFunction(hFile, CLRRTS);
end;

procedure THICOMMultiTh._work_doByteSize;
begin
  SetByteSize(ToInteger(_Data)); 
end;

procedure THICOMMultiTh._work_doStopBits;
begin
  SetStopBits(ToInteger(_Data));
end;

procedure THICOMMultiTh._work_doParity;
begin
  SetParity(ToInteger(_Data));
end;

procedure THICOMMultiTh._work_doReadBuffer;
begin
  _ReadBuffer := ToInteger(_Data);
end; 

procedure THICOMMultiTh._work_doInOutQueue;
begin
  _InOutQueue := ToInteger(_Data);
end; 

procedure THICOMMultiTh.SetByteSize;
begin
  _ByteSize := max(4, min(8, Value + 4));
end;

procedure THICOMMultiTh.SetStopBits;
begin
  case Value of
    0: _StopBits := ONESTOPBIT;
    1: _StopBits := ONE5STOPBITS;
    2: _StopBits := ONE5STOPBITS
  else
    _StopBits := ONESTOPBIT;    
  end;  
end;

procedure THICOMMultiTh.SetParity;
begin
  case Value of
    0: _Parity := NOPARITY;
    1: _Parity := ODDPARITY;
    2: _Parity := EVENPARITY;
    3: _Parity := MARKPARITY;
    4: _Parity := SPACEPARITY
  else
    _Parity := NOPARITY;    
  end;  
end;
    
end.